val libdir = "smlnj-lib/"
fun uselib f = use (libdir ^ "Util/" ^ f);
       
use "lisp.lex.sml";
uselib "hash-table-rep.sml";
uselib "mono-hash-table-sig.sml";
uselib "word-hash-table.sml";
uselib "hash-string.sml";
open TextIO;
open LispTokens;
open UserDeclarations;

(* Internal datatype for both code and data.
 * The Procedure constructor consists of
 *    an environment
 *    a list of parameter names
 *    an optional code body (see assignment) *)
datatype sexp = Int of int
              | Symbol of sexp
              | Name of string
              | List of sexp list
              | Procedure of sexp WordHashTable.hash_table list
                             * sexp list
                             * sexp list option
              | Bool of bool
              | VOID
              | DOT;

exception TableException;
exception InterpretException of string;
exception ParseException of string;

(* gives a correct stringRepresentation for every sexp *)
fun stringRep (s) = raise InterpretException("stringRep not implemented");

(* makes it easy to throw a very normal error exception *)
fun exptExn (hvor, expected, SOME(got):sexp option) =
    InterpretException("[" ^ hvor ^ "]: expected " ^ expected ^ ", but got: " ^ stringRep(got))
  |  exptExn (hvor, expected, NONE) =
     InterpretException("[" ^ hvor ^ "]: expected " ^ expected ^ ", but got: None");

(* These three procedures handles the symboltable (environment) *)
fun scopeCreate () = WordHashTable.mkTable(30, TableException) : sexp WordHashTable.hash_table;

local 
    val hs = HashString.hashString;
in
fun scopeLookup (hd::envStack, name) = (case WordHashTable.find hd (hs name) of
                                           SOME(s) => s
                                         | NONE => scopeLookup(envStack, name))
  | scopeLookup (nil,name) = raise InterpretException("scopeLookup: unable to lookup: " ^ name )
and scopeAdd (hd::envStack, name, value) = (WordHashTable.insert hd (hs(name), value); hd::envStack)
  | scopeAdd _ = raise InterpretException("shouldn't happen!")
end;

(* these procedure evaluates all the procedure primitives.
 * The sexp arguments in the last parameter are not evaluated
   at entry point *)
fun primitive (env, "eval", [arg]) = eval(env, eval(env, arg))
  | primitive (env, what, args) = raise InterpretException("primitive fail (not implemented): " ^ what ^ ": " ^ stringRep(hd(args)))

(* adds a mapping between the parameter names
 * and the arguments to the environment.
 * the sexp arguments should be evaluated before entry *)
and bindParamsToValues (env, pNames, args): sexp WordHashTable.hash_table list = 
    raise InterpretException("bindParamsToValues not implemented")

and evalList(env, l) = (map (fn n => eval(env, n)) l)

(* treats primitives and normal procedures different.
 * adds a new scope and makes sure that the evaluated
 * arguments are mapped to the parameter names *)
and procedureCall (env, Procedure(e, Name(fName)::a, NONE), args) = primitive(env, fName, args)
  | procedureCall(env, Procedure(pEnv, paramNames, SOME(body)), args)  = raise InterpretException("normal procedure calls not implemented ")
  | procedureCall _ = raise InterpretException("ProcedureCall: unknown call")

(* evaluates a sexp in relation to an environment *)
and eval (env, Symbol(s)) = s
  | eval (env, Name(s)) = scopeLookup(env, s)
  | eval (env, List(Name(fName)::args)) = procedureCall (env, scopeLookup(env, fName), args)
  | eval (env, List(Procedure(x)::args)) = procedureCall(env, Procedure(x), args)
  | eval (env, List(pProc::args)) = (case eval(env, pProc) of
                                         Procedure (a,b,c) => eval(env, List(Procedure(a,b,c)::args))
                                       | _ => raise exptExn("eval", "list with procedure as first element", SOME(List (pProc::args))))
  | eval (env, x) = x;

(* Tries to create a syntax tree from the stream of tokens *)
fun parse lexer = let
    fun pList (RPAREN):sexp list = []
      | pList (SYMBOL(s)) = [Name(s)] @ pList(lexer())
      | pList (NUM(n)) = [Int(n)] @ pList(lexer())
      | pList (QUOTE) = [p(QUOTE)] @ pList(lexer())
      | pList (LPAREN) = [List(pList(lexer()))] @ pList(lexer())
      | pList (BOOL(b)) = [Bool(b)] @ pList(lexer())
      | pList (D) = [DOT] @ pList(lexer())
      | pList _ = raise ParseException("Unable to parse list.")
    and
    p (LPAREN) = List(pList(lexer()))
  | p (NUM(n)) = Int(n)
  | p (QUOTE) =  Symbol(p(lexer()))
  | p (SYMBOL(s)) = Name(s)
  | p (BOOL(b)) = Bool(b)
  | p (EOF) = VOID
  | p (_) = raise ParseException("Unable to parse")
in p(lexer()) end;

(* add base procedures to library *)
val baseEnv = [scopeCreate()];
val primitives =  ["eval", "+", "-", "cons", "car", "cdr", "define", "=", "<", ">", "cond", "if", "null?", "pair?", "and", "or", "let", "remainder", "quotient","*", "lambda", "print"];
    map (fn n => scopeAdd(baseEnv, n, Procedure(baseEnv, [Name(n)], NONE))) primitives;

val lexer = makeLexer( fn n => (print "$ "; valOf(inputLine( stdIn )) ));

(* *)
fun itest() = stringRep (eval (baseEnv, parse(lexer)));
fun ptest() = parse(lexer);
fun etest() = eval (baseEnv, parse(lexer));
fun ltest() = lexer();

fun repl():unit = (print(itest() ^ "\n"); repl())
    handle InterpretException(s) => print("interpretException: " ^ s ^ "\n")
         | ParseException(s) => print("parseException: " ^ s ^ "\n");

fun readLib baseLib fileName = let

    fun addWhile (lexer) = (case parse(lexer) of
                                VOID => print("Library " ^ fileName ^ " loaded.\n")
                              | x => (eval([baseLib], x);
                                      addWhile (lexer)))
in
    let val s = openIn fileName in
        addWhile(makeLexer (fn n => inputN(s, n)))
        handle InterpretException(s) => print("interpretException: " ^ s ^ "\n")
             | ParseException(s) => print("parseException: " ^ s ^ "\n")
    end
end;

(* Reads the base library that is written in Scheme *)
readLib (hd baseEnv) "stdlib.scm";

(* When you think the interpreter is done,
 * uncomment to run tests. *)
(*
 readLib (hd baseEnv) "interpreter-tests.scm"; 
*)

(* Uncomment for prompt *)
(*
repl();
*)

